/*-----------------------------------------------------------------
  LED Alarm Clock Copyright (c) 2020 by Jack Botner

  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php

  led.c

  To display numbers on a 4 digit LED display. Since no segment
  encoder chip is used, we have to do the encoding of each digit
  ourself.
  
  Each LED segment maps to a bit on one of the ports.
  1=turn segment on, 0=turn segment off.

  20201212: program written
-----------------------------------------------------------------*/
#include <avr/io.h>
#include <ctype.h>
#include "common.h"
#include "led.h"

/*-----------------------------------------------------------------
  ledOutputData(...)

  Outputs a letter or number to the led display.

  Input: 1. which digit, 1..4
            This parameter is not needed in the sense that the
            digit is not selected here. It is used to blank out 
			the high-order hours digit when zero.
		 2. the number to show - 0..9 (anything else shows blank)
		 3. whether to show the decimal point. 0 = no DP, 1 = show DP
-----------------------------------------------------------------*/
void ledOutputData( uint8_t ucDigit, uint8_t ucNumber, uint8_t ucDp )
{
  // Segment A = bit 0, and so on, for each digit 0..9
  static const uint8_t ucDigTab[10] =
  {
    0b00111111, 0b00000110, 0b01011011, 0b01001111, 0b01100110,
    0b01101101, 0b01111101, 0b00000111, 0b01111111, 0b01100111
  };
  
  static const uint8_t ucLetterTab[6] =
  {
	0b01110111, 0b00111001, 0b00111000, 0b01101101,
    0b01111100, 0b01010000
  };

  static const uint8_t ucLetters[6] =
  {
	'A', 'C', 'L', 'S', 'B', 'R'
  };
  
  uint8_t nn, ucNibble, ucSegments, ucLetter;

  // Check for letter first
  if ( isalpha( ucNumber ) )
  {
	ucLetter = toupper( ucNumber );
    for ( nn=0; nn<6; ++nn )
    {
	  if ( ucLetter == ucLetters[nn] )
      {
	    ucSegments = ucLetterTab[nn];
        if ( ucDp )
          ucSegments |= 0x80;
        LED_SEG_PORT = ucSegments;
        return;
	  }
    }
	ucNumber = 0x0f;    // letter but unrecognized
  }

  ucNibble = ucNumber & 0x0f;
  ucSegments = ( ucNibble > 9 ) ? 0 : ucDigTab[ucNibble];

  // To blank out high-order hours digit when zero
  if ( ucDigit == 1 && !ucNumber ) ucSegments = 0;
  
  if ( ucDp )
    ucSegments |= 0x80;

  LED_SEG_PORT = ucSegments;
}

/*---------------------------------------------------------------------
  ledBlankDisplay(...)
  
  Turns all the digit selector transistors off.
---------------------------------------------------------------------*/
void ledBlankDisplay()
{
  clear_bit( LED_DIG_PORT, DIG_1_BIT );
  clear_bit( LED_DIG_PORT, DIG_2_BIT );
  clear_bit( LED_DIG_PORT, DIG_3_BIT );
  clear_bit( LED_DIG_PORT, DIG_4_BIT );
}

/*---------------------------------------------------------------------
  ledSelectDigit(...)
  
  Selects a digit of the led display, by turning on the requested
  driver transistor.
  
  Input: a single numeric digit, range 0..3. 0 drives the high order
         digit, 3 drives the low order digit. 
---------------------------------------------------------------------*/
void ledSelectDigit( uint8_t cDigit )
{
  if ( !cDigit )
    set_bit( LED_DIG_PORT, DIG_1_BIT );
  else
    if ( cDigit == 1 )
      set_bit( LED_DIG_PORT, DIG_2_BIT );
    else
      if ( cDigit == 2 )
        set_bit( LED_DIG_PORT, DIG_3_BIT );
      else
        if ( cDigit == 3 )
          set_bit( LED_DIG_PORT, DIG_4_BIT );
}

