/*---------------------------------------------------------------------------
  LED Alarm Clock.c
 
  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php
 
  Alarm clock with 4 digit LED display
  
  Processor: ATmega164P
  Clock: external, 16MHz
  Display: Luckylight KW4-56NCWB-P-Y Quadruple Digit Numeric Display (CC) (white)
  Rotary Encoder: EC11-1B-10T (Changzhou Xinze Electronic Co.) (or CTS 290)
  RTC: NJU6355
  Speaker
  Photocell

  20201212: program written
  20210215: add speaker beep when turning alarm on or off
  20210316: fix incorrect change of am/pm when setting clock
  20210526: move encoder I/O code out of interrupt handler
  20211114: rework wdt implementation
 ---------------------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#define F_CPU 16000000UL			// 16 MHz
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "led.h"
#include "nju6355.h"

#define WANT_WDT

// You can measure Vcc and update the define below.
// Not really necessary for this application. May
// have a tiny effect on the photocell threshold.
#define Vcc_MV               5000

// This controls when the display switches from dim to bright (Mv).
#define PHOTOCELL_THRESHOLD  1000

#define STATE_INITIAL		1
#define STATE_PROGRAM		2
#define STATE_RUN	        3

// Global variables

static volatile uint8_t	ucSystemState=STATE_INITIAL;

volatile uint8_t ucEventFlags=0;                 // system event flags

volatile uint16_t ucSpkrToggles=0;    // so alarm does not sound too long

// To control the brightness of the display, this must be
// a value between 0 and LED_TIME_SLICES
// and must be a multiple of 4.
// 0 = display off, LED_TIME_SLICES = max brightness
volatile uint8_t ucLedTimeSliceLimit=LED_TIME_SLICES;
volatile uint8_t ucLed[4] = { 9, 9, 9, 9 };	     // led digit values
volatile uint8_t ucDp[4]  = { 0, 0, 0, 0 };	     // which decimals are on

// Note: hours 0..23 adjusted to be 1..12, and PM flag set
volatile uint8_t ucClockHours=0, ucClockMins=0;
volatile uint8_t ucAlarmHours=7, ucAlarmMins=0;
volatile uint8_t ucClockPM=0, ucAlarmPM=0;	    // 0=AM, 1=PM
volatile uint8_t ucAlarmState=ALARM_OFF; 	    // alarm state
volatile uint8_t ucSeconds=0;		            // clock seconds

MY_EEPROM_DATA stEepromVars;			// variables kept in EEPROM

extern volatile uint16_t  uiAdc;        // last adc sample

extern volatile RTCPARMS stRtc;         // rtc data

static volatile uint8_t  ucDebouncedState;		    // debounced state of the switches
static volatile uint8_t  ucPreviousState=0; 		// previous switch state
static volatile uint8_t  ucNrUps=0, ucNrDowns=0; 	// Encoder logic
static volatile uint16_t uiMsPressed;			    // how long encoder button pressed

// internal functions
static void check_alarm( void);
static void display_alarm_time_for_2_secs(void);
static void spkr_beep(void);
static void system_init(void);
static void read_rtc(void);

/*---------------------------------------------------------------------------
  main()
---------------------------------------------------------------------------*/
int main(void)
{
  static volatile uint8_t ucTctr=0;		// update time counter
  static volatile uint8_t ucTctr2=0;	// update time counter2
  uint16_t    uiADCmv;
  uint8_t     ii;

  while (1)
  {
    switch ( ucSystemState )
    {
      case STATE_INITIAL:
        system_init();                  // initialize cpu
        for ( ii=0; ii<200; ++ii )      // wait a while
          _delay_loop_2( 0 );           // 16ms delay

        for ( ii=0; ii<4; ++ii )       // reset display
          ucLed[ii] = ucDp[ii] = 0;

        rtc_init();

        // To adjust clock using frequency counter, temporarily enable following
        // statement, load and run. Connect counter to TP1. Adjust trimmer C4
        // to 128.000 Hz. This is documented in the NJU6355 Application Note.
        //rtc_test();
          
        // Get the saved state variables from eeprom
        eeprom_read_block( &stEepromVars, 0, sizeof( MY_EEPROM_DATA ) );
        if ( stEepromVars.ucAlarmHrs > 12 || stEepromVars.ucAlarmMns > 59 || 
             !stEepromVars.ucAlarmHrs ||
             stEepromVars.ucAlrmPm > 1   || stEepromVars.ucAlarm_State > ALARM_ON_ACTIVE )
        {
          // Initialize all eeprom variables
          stEepromVars.ucAlarmHrs = ucAlarmHours;
          stEepromVars.ucAlarmMns = ucAlarmMins;
          stEepromVars.ucAlrmPm   = ucAlarmPM;
          stEepromVars.ucAlarm_State = ALARM_OFF;
          set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
        }
        else
        {
          // Initialize clock with saved eeprom variables
          ucAlarmHours = stEepromVars.ucAlarmHrs;
          ucAlarmMins  = stEepromVars.ucAlarmMns;
          ucAlarmPM    = stEepromVars.ucAlrmPm;
          ucAlarmState = stEepromVars.ucAlarm_State;
          if ( ucAlarmState != ALARM_OFF )
            ucDp[0] = 1;
        }

#ifdef WANT_WDT         
        wdt_enable( WDTO_1S );
#endif

        ucSystemState = STATE_RUN;
        break;

      case STATE_RUN:
#ifdef WANT_WDT
        wdt_reset();
#endif
        if ( test_bit( ucEventFlags, EVENT_ADC_COMPLETE ) )
        {
          // uiAdc0 (adresh:adresl) is in the range 0..1023 which represents a 
          // voltage of 0..Vcc volts (Vref). Scale the adc value to a voltage value.

          cli();
          uiADCmv = (unsigned int) ( ( (unsigned long) Vcc_MV * (unsigned long) uiAdc ) / 1024L );
          if ( uiADCmv > PHOTOCELL_THRESHOLD + 10 )
            ucLedTimeSliceLimit = LED_TIME_SLICES;
          else
            if ( uiADCmv < PHOTOCELL_THRESHOLD - 10 )
              ucLedTimeSliceLimit = 4;
          sei();
          clear_bit( ucEventFlags, EVENT_ADC_COMPLETE );
        }

        if ( test_bit( ucEventFlags, EVENT_1MS ) )
        {
          encoder();      // check for encoder I/O
          clear_bit( ucEventFlags, EVENT_1MS ); 
        }

        if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
        {
          if ( ucAlarmState == ALARM_ON_ACTIVE )
          {
            // If the alarm is sounding, just shut it up
            ucAlarmState = ALARM_ON_QUIET;
            set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
          }
          else
            ucSystemState = STATE_PROGRAM;
              
          clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
        }

        if ( test_bit( ucEventFlags, EVENT_ENCODER_CW ) ||
             test_bit( ucEventFlags, EVENT_ENCODER_CCW ) )
        {
          if ( ucAlarmState == ALARM_ON_ACTIVE )
          {
            // If the alarm is sounding, just shut it up
            ucAlarmState = ALARM_ON_QUIET;
            set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
          }
          else
          {
            // If alarm state is off, turn it on
            if ( ucAlarmState == ALARM_OFF )
            {
              ucAlarmState = ALARM_ON_QUIET;
              set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
              ucDp[0] = 1;
              display_alarm_time_for_2_secs();
            }
            else
              // If alarm is on, turn it off
              if ( ucAlarmState == ALARM_ON_QUIET )
              {
                ucAlarmState = ALARM_OFF;
                set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
                ucDp[0] = 0;
                spkr_beep();     // make a little noise
              }
              else
              {
                // Alarm is sounding, shut up
                ucAlarmState = ALARM_ON_QUIET;
                set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
              }
          }
              
          clear_bit( ucEventFlags, EVENT_ENCODER_CW );
          clear_bit( ucEventFlags, EVENT_ENCODER_CCW );
        }
          
        if ( test_bit( ucEventFlags, EVENT_HALFSEC ) )
        {
          if ( ucTctr )
          {
            // Update the time periodically
            read_rtc();
            display_led( ucClockHours, ucClockMins, ucClockPM );
            ucTctr = 0;
          }
          else
          {
            // Check the alarm periodically
            check_alarm();
            ucTctr = 1;
          }
              
          clear_bit( ucEventFlags, EVENT_HALFSEC );
        }
         
        ++ucTctr2;
        if ( ucTctr2 > 9 ) 
        {
          // check the ambient light
          set_bit( ADCSRA, ADSC );		// initiate a conversion
          ucTctr2 = 0;
        }

        if ( test_bit( ucEventFlags, EVENT_UPDATE_EEPROM ) )
        {
          stEepromVars.ucAlarmHrs = ucAlarmHours;
          stEepromVars.ucAlarmMns = ucAlarmMins;
          stEepromVars.ucAlrmPm   = ucAlarmPM;
          stEepromVars.ucAlarm_State = ucAlarmState;
          eeprom_write_block( &stEepromVars, 0, sizeof( MY_EEPROM_DATA ) );
          clear_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
        }
          
        break;

      case STATE_PROGRAM:
#ifdef WANT_WDT
        wdt_disable();
#endif
        program_clock();
#ifdef WANT_WDT
        wdt_enable( WDTO_1S );
#endif
        ucSystemState = STATE_RUN;
        break;
    }
  }
}

/*-----------------------------------------------------------------
  system_init()

  PORTA: PA0 (input)  Rotary Encoder "A"
		 PA1 (input)  Rotary Encoder "B"
		 PA2 (input)  Rotary Encoder "SW"
		 PA3 (input)  not used
		 PA4 (output) LED Digit 1 driver
		 PA5 (output) LED Digit 2 driver
		 PA6 (output) LED Digit 3 driver
		 PA7 (output) LED Digit 4 driver

  PORTB: PB0 (input)  not used
         PB1 (input)  not used
         PB2 (input)  not used
         PB3 (input)  not used
         PB4 (input)  not used
         PB5 (input)  MOSI ISP
         PB6 (input)  MISO ISP
         PB7 (input)  SCK ISP

  PORTC: PC0 (output) RTC Data
         PC1 (output) RTC I/O
         PC2 (output) RTC Clock
         PC3 (output) RTC CE
         PC4 (output) buzzer
         PC5 (output) LED colon
         PC6 (input)  not used
		 PC7 (input)  not used

  PORTD: PD0 (output) LED "a" segment
         PD1 (output) LED "b" segment
         PD2 (output) LED "c" segment
         PD3 (output) LED "d" segment
         PD4 (output) LED "e" segment
         PD5 (output) LED "f" segment
         PD6 (output) LED "g" segment
		 PD7 (output) LED "DP" segment
----------------------------------------------------------------*/
static void system_init()
{
  // Disable the WDT at startup
  MCUSR = 0;
  wdt_disable();

  // Set up the I/O ports directions, pull-ups
  DDRA  = 0b11110000;       // PortA0..3=in, 4-7=out
  PORTA = 0b00000111;       // initialize PortA pullups

  DDRB  = 0b00000000;       // PortB0-7=in
  PORTB = 0b11111111;       // initialize PortB pullups

  DDRC  = 0b00111111;       // PortC0-5=out, 6-7=in
  PORTC = 0b11000000;       // initialize PortC pullups

  DDRD  = 0b11111111;       // PortD0..7 = out
  PORTD = 0b00000000;       // initialize PortD

  PRR = 0b10010110;	    // power save: using timers & ADC only

  // Initialize timer0 for button debouncing
  TCNT0 = INIT_CTR0;
  TCCR0A = 0b00000000;       // timer/counter normal mode
  TCCR0B = 0b00000100;       // prescaler = 256 (turns timer0 on)
  set_bit( TIMSK0, TOIE0 );	 // enable timer/counter0 overflow interrupt

  // Initialize timer1 for 1/2 second timing
  TCNT1 = INIT_CTR1;
  TCCR1A = 0b00000000;		// normal operation
  TCCR1B = 0b00000100;		// prescaler = 256 (turns timer1 on)
  TCCR1C = 0b00000000;		// normal operation
  set_bit( TIMSK1, TOIE1 );	// enable timer/counter1 overflow interrupt

  // Initialize timer2 for LED management
  TCNT2 = INIT_CTR2;
  TCCR2A = 0b00000000;       // timer/counter normal mode
  TCCR2B = 0b00000011;       // prescaler = 32 (turns timer2 on)
  set_bit( TIMSK2, TOIE2 );	 // enable timer/counter0 overflow interrupt

  // Set up PA3/ADC3 as analog input, Vref=Vcc
  ADMUX  = 0b01000011;		// Vref=Vcc; result right adjusted; select ADC3
  ADCSRA = 0b10001111;		// Enable ADC; enable interrupt; prescaler = 128

  sei();                // enable global interrupts
}

/*-----------------------------------------------------------------
  check_alarm()
  
  If the alarm is on, compares the alarm time to the clock time.
  If match, sounds the alarm. Seconds are checked to prevent the
  alarm from sounding again after being silenced.
-----------------------------------------------------------------*/
static void check_alarm()
{
  if ( ucAlarmState == ALARM_ON_QUIET )
  {
    if ( ucClockHours == ucAlarmHours && ucClockMins == ucAlarmMins &&
	     ucClockPM == ucAlarmPM && ucSeconds < 2 )
    {
	  ucAlarmState = ALARM_ON_ACTIVE;
	  set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
    }
  }
  else
    if ( ucAlarmState == ALARM_ON_ACTIVE )
    {
      // This code to make sure alarm does not sound for more than a minute.
      if ( ucSpkrToggles > 30000 )
      {
        ucAlarmState = ALARM_ON_QUIET;
        set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
      }
    }
}

/*-----------------------------------------------------------------
  read_rtc()
  
  Get the time from the RTC.
  Adjust hours to show 12 hour format.
  Also saves seconds for alarm logic.
-----------------------------------------------------------------*/
static void read_rtc()
{
  uint8_t ucHr;

  rtc_read();    // information read into stRtc
  ucHr = stRtc.rtc_hours;
  // Since the clock is in 24 hour format (0..23), make adjustments
  ucClockPM = ( ucHr > 11 ) ? 1 : 0;

  if ( !ucHr )
  ucHr = 12;
    else
  if ( ucHr > 12 )
    ucHr = ucHr - 12;

  ucClockHours = ucHr;
  ucClockMins = stRtc.rtc_mins;
  ucSeconds = stRtc.rtc_secs;
}

/*-----------------------------------------------------------------
  display_led()
  
  Put the latest clock or alarm time on the led.
  It is possible to display blank hours or minutes by
  passing a value out of bounds.
-----------------------------------------------------------------*/
void display_led( uint8_t ucH, uint8_t ucM, uint8_t ucP )
{
  if ( ucH > 12 )
  {
    ucLed[0] = 0x0f;
    ucLed[1] = 0x0f;
  }
  else
  {
    ucLed[0] = ucH / 10;
    ucLed[1] = ucH % 10;
  }

  if ( ucM > 59 )
  {
    ucLed[2] = 0x0f;
    ucLed[3] = 0x0f;
  }
  else
  {
    ucLed[2] = ucM / 10;
    ucLed[3] = ucM % 10;
  }

  ucDp[3]  = ( ucP ) ? 1 : 0;
}

/*-----------------------------------------------------------------
  display_alarm_time_for_2_secs()

  The user just turned the alarm on, so show the alarm time for
  a couple of seconds. Also make the speaker bleep so the user
  knows the alarm has been activated.
-----------------------------------------------------------------*/
static void display_alarm_time_for_2_secs()
{
  uint8_t   ii, ucAlarmSave;

  display_led( ucAlarmHours, ucAlarmMins, ucAlarmPM );
  set_bit( COLON_PORT, COLON_PIN );

  ucAlarmSave = ucAlarmState;
  ucAlarmState = ALARM_ON_ACTIVE;
#ifdef WANT_WDT
  wdt_disable();
#endif

  for ( ii=0; ii<80; ++ii )      // wait 1.31 seconds
  {
    _delay_loop_2( 0 );          // 16.38ms delay

    if ( ii == 2 )
      // This turns the beeper off after 32 ms
      ucAlarmState = ucAlarmSave;
  }

#ifdef WANT_WDT
  wdt_enable( WDTO_1S );
#endif
}

/*-----------------------------------------------------------------
  spkr_beep()

  Sound the speaker for a bit.
-----------------------------------------------------------------*/
static void spkr_beep()
{
  uint8_t   ucAlarmSave;

  ucAlarmSave = ucAlarmState;
  ucAlarmState = ALARM_ON_ACTIVE;

  _delay_loop_2( 0 );          // 16.38ms delay
  _delay_loop_2( 0 );          // 16.38ms delay

  ucAlarmState = ucAlarmSave;
}
