/*-----------------------------------------------------------------
  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php

  led alarm clock by Jack Botner

  program.c - clock programming routines

  To program the clock's time or the alarm time.
  
  20201212 - program written.
  20210526: move encoder I/O code out of interrupt handler
-----------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/wdt.h>
#include "main.h"
#include "nju6355.h"

// Global variables
extern volatile uint8_t  ucEventFlags;		    // system event flags
extern volatile RTCPARMS stRtc;                 // rtc data

// Note: hours 0..23 adjusted to be 1..12, and PM flag set
extern volatile uint8_t ucClockHours, ucClockMins;
extern volatile uint8_t ucAlarmHours, ucAlarmMins;
extern volatile uint8_t  ucAlarmState;	 	    // alarm state
extern volatile uint8_t  ucClockPM, ucAlarmPM;	// 0=AM, 1=PM
extern volatile uint8_t  ucSeconds;

extern volatile uint8_t ucLed[4];			    // led digit values
extern volatile uint8_t ucDp[4];			    // which decimals are on

#define PROGRAM_CLOCK     1
#define PROGRAM_ALARM     2

// Internal functions
static uint8_t get_encoder_event( void );
static void    program_clock_time( uint8_t );
static void    update_rtc(void);

/*-----------------------------------------------------------------------
  program_clock()
  
  Programs the clock times and display brightness.
  The rotary encoder is used to invoke menus and selections.

  1. Press the encoder button. Set alarm menu appears. If the button
     is pressed again, the hours are displayed and can be adjusted by
     turning the encoder. Another button press and the minutes can be
     adjusted. Another button press returns to normal operation.
     If you don't want to program the alarm time, turn the encoder
     and you go on the programming the clock time.
  2. Programming the alarm time is exactly the same as the clock time.
-----------------------------------------------------------------------*/
void program_clock()
{
  uint8_t ucEncoderEvent;

  clear_bit( ucEventFlags, EVENT_ENCODER_CW );
  clear_bit( ucEventFlags, EVENT_ENCODER_CCW );
  clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );

  // Show 'AL'
  ucLed[0] = 'A';
  ucLed[1] = 'L';
  ucLed[2] = ucLed[3] = 0x0f;

  ucEncoderEvent = get_encoder_event();

  // User may press button meaning program the alarm time.
  // Otherwise a turn of the encoder means go on to next step.

  if ( ucEncoderEvent == EVENT_BUTTON_PRESSED )
  {
    program_clock_time( PROGRAM_ALARM );
    clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
    return;
  }

  // Show 'CL'
  ucLed[0] = 'C';
  ucLed[1] = 'L';
  ucLed[2] = ucLed[3] = 0x0f;

  ucEncoderEvent = get_encoder_event();

  // User may press button meaning program the clock time.
  // Otherwise a turn of the encoder means done.

  if ( ucEncoderEvent == EVENT_BUTTON_PRESSED )
  {
    program_clock_time( PROGRAM_CLOCK );
    clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
  }
}

/*-----------------------------------------------------------------
  program_clock_time(...)

  input = whether programming clock or alarm

  Display hours ... turning the knob changes hours.
  Press button ... display minutes ... turning the knob changes
  minutes. Press button ... return.
-----------------------------------------------------------------*/
static void program_clock_time( uint8_t uiProgramWhat )
{
  uint8_t  ucHours, ucMinutes, ucPM, ucEncoderEvent;
  uint8_t  ucHoursSaved, ucMinsSaved, ucPMsaved, ucChanged=0;

  if ( uiProgramWhat == PROGRAM_CLOCK )
  {
    // Program the clock.
    ucHours   = ucClockHours;
    ucMinutes = ucClockMins;
    ucPM      = ucClockPM;
  }
  else
  {
    // Program the alarm.
    ucHours   = ucAlarmHours;
    ucMinutes = ucAlarmMins;
    ucPM      = ucAlarmPM;
  }

  ucHoursSaved = ucHours;
  ucMinsSaved = ucMinutes;
  ucPMsaved   = ucPM;

  while ( 1 )
  {
    // Show hours
    display_led( ucHours, 0xff, ucPM );

    ucEncoderEvent = get_encoder_event();
    if ( ucEncoderEvent == EVENT_BUTTON_PRESSED )
      break;

    if ( ucEncoderEvent == EVENT_ENCODER_CW )
    {
      if ( ucHours == 11 )          // 20210316
        ucPM = ( ucPM ) ? 0 : 1;
      
      if ( ucHours == 12 )
        ucHours = 1;
      else
         ++ucHours;
    }

    if ( ucEncoderEvent == EVENT_ENCODER_CCW )
    {
      if ( ucHours == 12 )          // 20210316
        ucPM = ( ucPM ) ? 0 : 1;

      if ( ucHours == 1 )
        ucHours = 12;
      else
        --ucHours;
    }
  }

  while ( 1 )
  {
    // Show minutes
    display_led( 0xff, ucMinutes, ucPM );

    ucEncoderEvent = get_encoder_event();
    if ( ucEncoderEvent == EVENT_BUTTON_PRESSED )
      break;

    if ( ucEncoderEvent == EVENT_ENCODER_CW )
    {
      if ( ucMinutes == 59 )
        ucMinutes = 0;
      else
        ++ucMinutes;
    }

    if ( ucEncoderEvent == EVENT_ENCODER_CCW )
    {
      if ( !ucMinutes )
        ucMinutes = 59;
      else
        --ucMinutes;
    }
  }

  if ( ucHoursSaved != ucHours || ucMinsSaved != ucMinutes || ucPMsaved != ucPM )
    ucChanged = 1;

  if ( ucChanged )
  {
    if ( uiProgramWhat == PROGRAM_CLOCK )
    {
      // Update the clock.
      ucClockHours = ucHours;
      ucClockMins = ucMinutes;
      ucSeconds = 0;
      ucClockPM = ucPM;
      update_rtc();
    }
    else
    {
      // Update the alarm.
      ucAlarmHours = ucHours;
      ucAlarmMins = ucMinutes;
      ucAlarmPM = ucPM;
      set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
    }
  }  
}

/*-----------------------------------------------------------------
  update_rtc()
  
  Update the RTC.
-----------------------------------------------------------------*/
static void update_rtc()
{
  stRtc.rtc_hours = ucClockHours;
  if ( ucClockPM )
    stRtc.rtc_hours += 12;
  stRtc.rtc_mins = ucClockMins;
  stRtc.rtc_secs = 0;

  rtc_write();
}

/*-----------------------------------------------------------------
  get_encoder_event()

  Wait for the button to be pressed or the knob turned.
-----------------------------------------------------------------*/
static uint8_t get_encoder_event()
{
  while ( 1 )
  {
#ifdef WANT_WDT
    wdt_reset();
#endif
    if ( test_bit( ucEventFlags, EVENT_1MS ) )
    {
      encoder();      // check for encoder I/O
      clear_bit( ucEventFlags, EVENT_1MS );
    }

    if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
    {
      clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
      return EVENT_BUTTON_PRESSED;
    }

    if ( test_bit( ucEventFlags, EVENT_ENCODER_CW ) )
    {
      clear_bit( ucEventFlags, EVENT_ENCODER_CW );
      return EVENT_ENCODER_CW;
    }

    if ( test_bit( ucEventFlags, EVENT_ENCODER_CCW ) )
    {
      clear_bit( ucEventFlags, EVENT_ENCODER_CCW );
      return EVENT_ENCODER_CCW;
    }
  }
}